@@ -26,9 +26,12 @@
 struct near_p2p_driver {
 	const char *name;
 	const char *service_name;
+	const char *fallback_service_name;
 	near_bool_t (*read)(int client_fd,
 				uint32_t adapter_idx, uint32_t target_idx,
-				near_tag_io_cb cb);
+				near_device_io_cb cb);
+	int (*push)(int client_fd, uint32_t adapter_idx, uint32_t target_idx,
+			struct near_ndef_message *ndef, near_device_io_cb cb);
 	void (*close)(int client_fd, int err);
 };
 
@@ -37,8 +40,8 @@ struct near_p2p_driver {
 int npp_init(void);
 void npp_exit(void);
 
-int snep_init(void);
-void snep_exit(void);
+int snep_core_init(void);
+void snep_core_exit(void);
 
 int handover_init(void);
 void handover_exit(void);
